@@ -229,7 +229,7 @@ TEST_F(DStubControllerTest, missingConfigFileArgument) {
     int argc = 2;
 
     // Record start time, and invoke launch().
-    EXPECT_THROW(launch(argc, argv), ProcessInitError);
+    EXPECT_THROW(launch(argc, argv), LaunchError);
 }
 
 /// @brief Tests launch with an operational error during application execution.
@@ -242,9 +242,9 @@ TEST_F(DStubControllerTest, launchRuntimeError) {
     isc::asiolink::IntervalTimer timer(*getIOService());
     timer.setup(genFatalErrorCallback, 2000);
 
-    // Write the valid, empty, config and then run launch() for 1000 ms
+    // Write the valid, empty, config and then run launch() for 5000 ms
     time_duration elapsed_time;
-    EXPECT_THROW(runWithConfig("{}", 2000, elapsed_time), ProcessRunError);
+    EXPECT_THROW(runWithConfig("{}", 5000, elapsed_time), ProcessRunError);
 
     // Verify that duration of the run invocation is the same as the
     // timer duration.  This demonstrates that the shutdown was driven
