@@ -1,4 +1,4 @@
-// Copyright (C) 2014 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2014-2015 Internet Systems Consortium, Inc. ("ISC")
 //
 // Permission to use, copy, modify, and/or distribute this software for any
 // purpose with or without fee is hereby granted, provided that the above
@@ -16,8 +16,8 @@
 #define CFG_OPTION_H
 
 #include <dhcp/option.h>
+#include <dhcp/option_space_container.h>
 #include <dhcpsrv/key_from_key.h>
-#include <dhcpsrv/option_space_container.h>
 #include <boost/multi_index_container.hpp>
 #include <boost/multi_index/hashed_index.hpp>
 #include <boost/multi_index/sequenced_index.hpp>
@@ -26,6 +26,8 @@
 #include <boost/shared_ptr.hpp>
 #include <stdint.h>
 #include <string>
+#include <set>
+#include <list>
 
 namespace isc {
 namespace dhcp {
@@ -197,6 +199,14 @@ typedef OptionContainer::nth_index<2>::type OptionContainerPersistIndex;
 class CfgOption {
 public:
 
+    /// @brief default constructor
+    CfgOption();
+
+    /// @brief Indicates the object is empty
+    ///
+    /// @return true when the object is empty
+    bool empty() const;
+
     /// @name Methods and operators used for comparing objects.
     ///
     //@{
@@ -259,6 +269,8 @@ public:
     /// as a parameter. If an item exists in the destination it is not
     /// copied.
     ///
+    /// @note: this method is not longer used so should become private.
+    ///
     /// @param [out] other Configuration object to merge to.
     void mergeTo(CfgOption& other) const;
 
@@ -403,6 +415,9 @@ typedef boost::shared_ptr<CfgOption> CfgOptionPtr;
 /// @brief Const pointer.
 typedef boost::shared_ptr<const CfgOption> ConstCfgOptionPtr;
 
+/// @brief Const pointer list.
+typedef std::list<ConstCfgOptionPtr> CfgOptionList;
+
 //@}
 
 }
