@@ -657,9 +657,8 @@ MasterLoader::MasterLoaderImpl::generateForIter(const std::string& str,
               continue;
           }
 
-          // 'it' can be equal to str.end() here, but it is handled
-          // correctly.
-          if (*it != '{') {
+          // The str.end() check is required.
+          if ((it == str.end()) || (*it != '{')) {
               // There is no modifier (between {}), so just copy the
               // passed number into the generated string.
               rstr += boost::str(boost::format("%d") % num);
@@ -671,7 +670,7 @@ MasterLoader::MasterLoaderImpl::generateForIter(const std::string& str,
               int offset = 0;
               unsigned int width;
               char base[2] = {'d', 0}; // char plus null byte
-              // cppcheck-suppress invalidscanf
+              // cppcheck-suppress invalidscanf_libc
               const int n = sscanf(scan_str, "{%d,%u,%1[doxXnN]}",
                                    &offset, &width, base);
               switch (n) {
@@ -808,7 +807,7 @@ MasterLoader::MasterLoaderImpl::doGenerate() {
     unsigned int start;
     unsigned int stop;
     unsigned int step;
-    // cppcheck-suppress invalidscanf
+    // cppcheck-suppress invalidscanf_libc
     const int n = sscanf(range.c_str(), "%u-%u/%u", &start, &stop, &step);
     if ((n < 2) || (stop < start)) {
         reportError(lexer_.getSourceName(), lexer_.getSourceLine(),
