@@ -1,4 +1,4 @@
-// Copyright (C) 2011  Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2011, 2015  Internet Systems Consortium, Inc. ("ISC")
 //
 // Permission to use, copy, modify, and/or distribute this software for any
 // purpose with or without fee is hereby granted, provided that the above
@@ -176,7 +176,7 @@ protected:
     {
         std::string unix_file = getSocketPath();
 
-        unlink(unix_file.c_str());
+        static_cast<void>(remove(unix_file.c_str()));
         test_un_.sun_family = AF_UNIX;
         strncpy(test_un_.sun_path, unix_file.c_str(), sizeof(test_un_.sun_path));
 #ifdef HAVE_SA_LEN
@@ -188,7 +188,7 @@ protected:
         if (listen_fd_ != -1) {
             close(listen_fd_);
         }
-        unlink(getSocketPath().c_str());
+        static_cast<void>(remove(getSocketPath().c_str()));
     }
 
     // Start an internal "socket session server".
@@ -201,7 +201,7 @@ protected:
             isc_throw(isc::Unexpected, "failed to create UNIX domain socket" <<
                       strerror(errno));
         }
-        if (bind(listen_fd_, convertSockAddr(&test_un_), test_un_len_) == -1) {
+        if (::bind(listen_fd_, convertSockAddr(&test_un_), test_un_len_) == -1) {
             isc_throw(isc::Unexpected, "failed to bind UNIX domain socket" <<
                       strerror(errno));
         }
